"""
Configure Django REST framework and drf-yasg.

This requires the `django-oauth-toolkit` and `drf-yasg` packages to be installed.
"""

from __future__ import annotations

from typing import Any

# When SessionAuthentication is allowed, it's critical that the following settings
# (respectively part of Django and django-cors-headers) are set to these values (although those
# are the also the default values).
SESSION_COOKIE_SAMESITE = "Lax"
CORS_ALLOW_CREDENTIALS = False

REST_FRAMEWORK: dict[str, Any] = {
    "DEFAULT_AUTHENTICATION_CLASSES": [
        "oauth2_provider.contrib.rest_framework.OAuth2Authentication",
        # Allow SessionAuthentication, as this is much more convenient for Ajax requests
        # from server-rendered pages, including:
        # * YASG (Swagger / ReDoc)
        # * The Admin interface, when using interactive fields like S3-file-field
        # * Augmentation of server-rendered views with background Javascript
        #   (see https://docs.djangoproject.com/en/3.1/ref/csrf/#ajax )
        # It's important that true SPAs and other clients be forced to go though
        # OAuth2Authentication instead, as this is the only supported auth mechanism which
        # robustly works across origins; however, it turns out that this can only be enforced
        # partially.
        # To understand why, first read https://web.dev/same-site-same-origin/ to understand
        # that even with "SameSite=Lax" (or "SameSite=Strict"), cookies are only technically
        # limited to same-site requests, and do not have the stronger same-origin limitation.
        # If a naive SPA developer configures their client to include credentials
        # ("{withCredentials: true}" in XHR, jQuery, and Axios, or "{credentials: 'include'}"
        # in Fetch; this configuration is often suggested as an "easy" fix for authentication
        # problems by StackOverflow), then the session cookie will be sent with any cross-site
        # requests where the user has logged into the Django server. Note, cross-site requests
        # include origins with a different port (typical in local development) and origins with
        # a different subdomain (common in many deployments). From DRF's perspective, as long
        # as the request uses a safe verb (more on this below), the request will be
        # authenticated transparently. However, since Django is configured to not set
        # "Access-Control-Allow-Credentials", the SPA client will not be able to read the
        # response and get a CORS error; this is the right outcome (client cannot effectively
        # make the request), but with a confusing and hard-to-debug reason (CORS error,
        # instead of a 401/403) and developers may be tempted to "fix" it by enabling
        # "Access-Control-Allow-Credentials" instead of fixing their client to use OAuth
        # correctly, which will likely lead to further bugs when the SPA is deployed to a
        # non-same-site environment. Alternatively, if the request does not use a safe verb
        # (and the request is not preflighted by the browser, which is permitted in some
        # cases), DRF will enforce CSRF protection and the request will 403 fail with a
        # "CSRF Failed: CSRF token missing or incorrect" message, which is also confusing and
        # may lead developers to incorrect fixes.
        # TL;DR: Developers of SPAs may encounter misleading error messages when making Ajax
        # requests "withCredentials", but security is still maintained.
        "rest_framework.authentication.SessionAuthentication",
    ],
    # This is a much more sensible degree of basic security
    "DEFAULT_PERMISSION_CLASSES": ["rest_framework.permissions.IsAuthenticatedOrReadOnly"],
    # BoundedLimitOffsetPagination provides LimitOffsetPagination with a maximum page size
    "DEFAULT_PAGINATION_CLASS": "resonant_utils.rest_framework.BoundedLimitOffsetPagination",
    # This provides a sane default for requests that do not specify a page size.
    # This also ensures that endpoints with pagination will always return a
    # pagination-structured response.
    "PAGE_SIZE": 100,
    # Real clients typically JSON-encode their request bodies, so the test client should too
    "TEST_REQUEST_DEFAULT_FORMAT": "json",
}

SWAGGER_SETTINGS: dict[str, Any] = {
    # The default security definition ("basic") is not supported by this DRF configuration,
    # so expect all logins to come via the Django session, which there's no OpenAPI
    # security definition for.
    "SECURITY_DEFINITIONS": None,
    "USE_SESSION_AUTH": True,
    # Needed because https://github.com/axnsan12/drf-yasg/pull/911
    "USE_COMPAT_RENDERERS": False,
}

REDOC_SETTINGS: dict[str, Any] = {}

__all__ = [
    "CORS_ALLOW_CREDENTIALS",
    "REDOC_SETTINGS",
    "REST_FRAMEWORK",
    "SESSION_COOKIE_SAMESITE",
    "SWAGGER_SETTINGS",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar

from django.contrib.auth import get_user_model
from django.contrib.auth.management.commands import createsuperuser

from resonant_settings.allauth_support.utils import temporarily_change_attributes

if TYPE_CHECKING:
    from django.contrib.auth.models import User, UserManager
else:
    User = get_user_model()
    UserManager = User._default_manager.__class__


class Command(createsuperuser.Command):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.UserModel = EmailAsUsernameProxyUser
        self.username_field = self.UserModel._meta.get_field(self.UserModel.USERNAME_FIELD)

    def _validate_username(
        self, username: str, verbose_field_name: str, database: str
    ) -> str | None:
        # Since "username" is actually unique, "email" (i.e. "self.username_field") is logically
        # unique too. Explicitly setting the "_unique" attribute ensures that app-level duplicate
        # checking is done by "_validate_username", which produces better, earlier error messages.
        # Out of an abundance of caution, set the "_unique" attribute back to its original value
        # when this is done.
        with temporarily_change_attributes(self.username_field, _unique=True):
            # Normalize (as it would be done before saving) for better duplicate detection
            username = self.UserModel.normalize_username(username)
            return super()._validate_username(  # type: ignore[misc,no-any-return]
                username, verbose_field_name, database
            )


class EmailAsUsernameProxyUserManager(UserManager["EmailAsUsernameProxyUser"]):
    # This version of "create_superuser" makes the "username" argument optional
    def create_superuser(
        self,
        username: str | None = None,
        email: str | None = None,
        password: str | None = None,
        **extra_fields: Any,
    ) -> EmailAsUsernameProxyUser:
        # Practically, email will always be provided
        if email is None:
            raise ValueError("Email address must be provided.")
        return super().create_superuser(
            username=email, email=email, password=password, **extra_fields
        )


class EmailAsUsernameProxyUser(User):
    class Meta(User.Meta):
        proxy = True

    objects = EmailAsUsernameProxyUserManager()

    # "createsuperuser.Command" automatically includes the referent of "USERNAME_FIELD", and we want
    # to apply username labeling, help text, and validation rules from the actual "email" field
    USERNAME_FIELD = "email"

    # Don't include "email" in "REQUIRED_FIELDS", to prevent adding that field twice to the
    # "createsuperuser.Command" argument parser
    REQUIRED_FIELDS: ClassVar[list[str]] = []

    @classmethod
    def normalize_username(cls, username: str) -> str:
        # This method is called from "UserManager._create_user" with the actual "username" field.
        # To ensure that the saved value of the "username" field exactly matches the "email" field,
        # apply the same normalization process.
        return EmailAsUsernameProxyUserManager.normalize_email(username)



from __future__ import annotations

from resonant_settings._env import env

# Acknowledge early in development, which will help prevent failing or
# long-running tasks from being started automatically every time the worker
# process restarts; this more aggressively flushes the task queue.
CELERY_TASK_ACKS_LATE = False

CELERY_TASK_ALWAYS_EAGER: bool = env.bool("DJANGO_CELERY_TASK_ALWAYS_EAGER", default=False)
# In eager mode (which might be set directly in tests), non-propagated exceptions allow bugs to go
# unnoticed, so ensure this is always enabled. This should have no effect in non-eager mode.
CELERY_TASK_EAGER_PROPAGATES = True

# In development, run without concurrency.
CELERY_WORKER_CONCURRENCY: int | None = 1

__all__ = [
    "CELERY_TASK_ACKS_LATE",
    "CELERY_TASK_ALWAYS_EAGER",
    "CELERY_TASK_EAGER_PROPAGATES",
    "CELERY_WORKER_CONCURRENCY",
]



"""
Configure Django OAuth Toolkit.

This provides the following features:
* Harden security
* Improve usability of token scopes
* Improve quality of live for out of band flows and non-refreshing clients

This requires the `django-oauth-toolkit` package to be installed.
"""

from __future__ import annotations

from datetime import timedelta
from typing import Any

OAUTH2_PROVIDER: dict[str, Any] = {
    "ALLOWED_REDIRECT_URI_SCHEMES": ["https"],
    # This doesn't immediately change behavior, it just authorizes the use of wildcards in
    # Application redirect URIs, which is necessary to support SPA branch previews
    "ALLOW_URI_WILDCARDS": True,
    # Don't require users to re-approve scopes each time
    "REQUEST_APPROVAL_PROMPT": "auto",
    # ERROR_RESPONSE_WITH_SCOPES is only used with the "permission_classes" helpers for scopes.
    # If the scope itself is confidential, this could leak information. but the usability
    # benefit is probably worth it.
    "ERROR_RESPONSE_WITH_SCOPES": True,
    # Allow 5 minutes for a flow to exchange an auth code for a token. This is typically
    # 60 seconds but out-of-band flows may take a bit longer. A maximum of 10 minutes is
    # recommended: https://datatracker.ietf.org/doc/html/rfc6749#section-4.1.2.
    "AUTHORIZATION_CODE_EXPIRE_SECONDS": timedelta(minutes=5).total_seconds(),
    # Django can persist logins for longer than this via cookies,
    # but non-refreshing clients will need to redirect to Django's auth every 24 hours.
    "ACCESS_TOKEN_EXPIRE_SECONDS": timedelta(days=1).total_seconds(),
    # This allows refresh tokens to eventually be removed from the database by
    # "manage.py cleartokens". This value is not actually enforced when refresh tokens are
    # checked, but it can be assumed that all clients will need to redirect to Django's auth
    # every 30 days.
    "REFRESH_TOKEN_EXPIRE_SECONDS": timedelta(days=30).total_seconds(),
}

__all__ = [
    "OAUTH2_PROVIDER",
]



"""Configure a basic Django project."""

from __future__ import annotations

from typing import Any

TEMPLATES: list[dict[str, Any]] = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

PASSWORD_HASHERS: list[str] = [
    # Argon2 is recommended by OWASP, so make it the default for new passwords
    # https://cheatsheetseries.owasp.org/cheatsheets/Password_Storage_Cheat_Sheet.html
    "django.contrib.auth.hashers.Argon2PasswordHasher",
    # scrypt was the default hasher in older versions of Resonant,
    # so it must be enabled to read old passwords
    "django.contrib.auth.hashers.ScryptPasswordHasher",
    # Support for other hashers isn't needed,
    # since databases shouldn't have entries with other algorithms
]
# https://docs.djangoproject.com/en/stable/ref/settings/#auth-password-validators
AUTH_PASSWORD_VALIDATORS: list[dict[str, str]] = [
    {"NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"},
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"},
]

__all__ = [
    "AUTH_PASSWORD_VALIDATORS",
    "PASSWORD_HASHERS",
    "TEMPLATES",
]



"""
Configure S3Storage.

The following environment variables must be externally set:
* AWS_DEFAULT_REGION
* AWS_ACCESS_KEY_ID
* AWS_SECRET_ACCESS_KEY
* DJANGO_STORAGE_BUCKET_NAME

This requires the `django-storages[s3]` package to be installed.
"""

from __future__ import annotations

from datetime import timedelta

from resonant_settings._env import env

# These exact environment variable names are important,
# as direct instantiations of Boto will also respect them.
AWS_S3_REGION_NAME: str = env.str("AWS_DEFAULT_REGION")
AWS_S3_ACCESS_KEY_ID: str = env.str("AWS_ACCESS_KEY_ID")
AWS_S3_SECRET_ACCESS_KEY: str = env.str("AWS_SECRET_ACCESS_KEY")

AWS_STORAGE_BUCKET_NAME: str = env.str("DJANGO_STORAGE_BUCKET_NAME")

# It's critical to use the v4 signature;
# it isn't the upstream default only for backwards compatibility reasons.
AWS_S3_SIGNATURE_VERSION = "s3v4"

AWS_S3_MAX_MEMORY_SIZE = 5 * 1024 * 1024

# Although overwriting existing files can be dangerous, it's the application's responsibility
# (not the Storage layer) to handle. Setting this to `False` allows file names to be mutated in
# the event of a collision, which can be confusing.
AWS_S3_FILE_OVERWRITE = True
AWS_QUERYSTRING_EXPIRE = int(timedelta(hours=6).total_seconds())

__all__ = [
    "AWS_QUERYSTRING_EXPIRE",
    "AWS_S3_ACCESS_KEY_ID",
    "AWS_S3_FILE_OVERWRITE",
    "AWS_S3_MAX_MEMORY_SIZE",
    "AWS_S3_REGION_NAME",
    "AWS_S3_SECRET_ACCESS_KEY",
    "AWS_S3_SIGNATURE_VERSION",
    "AWS_STORAGE_BUCKET_NAME",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from allauth.account.adapter import DefaultAccountAdapter

if TYPE_CHECKING:
    from django.contrib.auth.models import AbstractUser
    from django.http import HttpRequest


class EmailAsUsernameAccountAdapter(DefaultAccountAdapter):  # type: ignore[misc]
    """Automatically populate the username as the email address."""

    def populate_username(self, request: HttpRequest, user: AbstractUser) -> None:
        user.username = user.email



"""
Configure Django Debug Toolbar.

This provides the following features:
* Improve performance with large queries

This requires the `django-debug-toolbar` package to be installed.
"""

from __future__ import annotations

from typing import Any

DEBUG_TOOLBAR_CONFIG: dict[str, Any] = {
    # The default size often is too small, causing an inability to view queries
    "RESULTS_CACHE_SIZE": 250,
    # If this setting is True, large sql queries can cause the page to render slowly
    "PRETTIFY_SQL": False,
}

__all__ = [
    "DEBUG_TOOLBAR_CONFIG",
]



from __future__ import annotations

from django.apps import AppConfig


class AllauthSupportConfig(AppConfig):
    name = "resonant_settings.allauth_support"
    verbose_name = "Resonant settings django-allauth support"
